"""
Original © NetApp 2024

A test module for the recline.repl.shell module
"""

import asyncio
import builtins

import pytest

import recline
from recline.repl import shell


@pytest.mark.parametrize("user_input, expected_marker, expected_output", [
    ("ut command -arg 2", 2, ""),
    ("ut command", None, "required: -arg"),
    ("ut command -arg 5", None, "This is a UT failure"),
    ("ut command -arg foo", None, "invalid int value"),
    ("bad command", None, "Unknown command"),
    ("ut command -arg 2 && ut command -arg 3", 3, ""),
    ("ut command; ut command -arg 3", 3, "required: -arg"),
    ("ut command -arg 5 && ut command -arg 2", None, "This is a UT failure"),
    ("ut command -arg 2 || ut command -arg 1", 2, ""),
    ("bad command; bad other command || ut command -arg 3", 3, "Unknown command"),
])
def test_shell_execute(user_input, expected_marker, expected_output, capsys):
    """Test that our shell can run one or more commands on input"""

    ut_marker = None

    @recline.command(name="ut command")
    def ut_command(arg: int):  # pylint: disable=unused-variable
        if arg == 5:
            raise ValueError("This is a UT failure")
        nonlocal ut_marker
        ut_marker = arg

    shell.execute(user_input)
    assert ut_marker == expected_marker
    captured = capsys.readouterr()
    assert expected_output in captured.out + captured.err


@pytest.mark.parametrize("user_input, expected_marker", [
    ("ut async command -arg 2", 2), ("ut async command -arg 30", 30),
])
def test_shell_execute_async_command(user_input, expected_marker):
    """Verify we can run async commands as well"""

    ut_marker = None

    @recline.command(name="ut async command")
    async def ut_command(arg: int):  # pylint: disable=unused-variable
        loops = 0
        while loops < arg:
            loops += 1
            await asyncio.sleep(0.001)
        nonlocal ut_marker
        ut_marker = arg

    shell.execute(user_input)
    assert ut_marker == expected_marker


def test_run_startup_exit_command(monkeypatch):
    """Verify that a command which is marked to run at startup or exit gets run"""

    startup_command_ran = False
    recline.commands.START_COMMAND = None

    def mock_eof(prompt):
        raise EOFError("UT is finished")

    monkeypatch.setattr(builtins, "input", mock_eof)

    @recline.command(atstart=True)
    def startup():  # pylint: disable=unused-variable
        nonlocal startup_command_ran
        startup_command_ran = True

    with pytest.raises(SystemExit):
        shell.relax(argv=["ut_program"])

    assert startup_command_ran
    recline.commands.START_COMMAND = None


@pytest.mark.parametrize("motd, expected", [
    ("This is a simple message", "This is a simple message"),
    (lambda: "This is a dynamic message", "This is a dynamic message"),
])
def test_run_motd(motd, expected, monkeypatch, capsys):
    """Verify the MOTD gets printed if one is provided"""

    def mock_eof(prompt):
        raise EOFError("UT is finished")

    monkeypatch.setattr(builtins, "input", mock_eof)
    with pytest.raises(SystemExit):
        shell.relax(argv=["ut_program"], motd=motd)

    captured = capsys.readouterr()
    assert expected in captured.out


def test_run_with_dash_c():
    """Verify only a single command is run when -c is passed in"""

    @recline.command(name="single command")
    def single_command():  # pylint: disable=unused-variable
        return 73

    assert shell.relax(argv=["ut_program", "-c", "single", "command"]) == 73


def test_run_non_repl():
    """Verify that if a program is not trying to be a repl, then we will parse
    a command from the input and exit
    """

    @recline.command(name="single command")
    def single_command():  # pylint: disable=unused-variable
        return 73

    assert shell.relax(argv=["ut_program", "single", "command"], repl_mode=False) == 73


def test_run_single_command():
    """Verify that if a program is not trying to be a repl, then we will parse
    a command from the input and exit
    """

    @recline.command(name="single command")
    def single_command():  # pylint: disable=unused-variable
        return 73

    assert shell.relax(argv=["ut_program"], single_command="single command") == 73



"""Common methods for parsing."""

import typing as T


class ParseError(RuntimeError):
    """Base class for all parsing related errors."""

    pass


class DocstringMeta:
    """
    Docstring meta information.

    Symbolizes lines in form of

        :param arg: description
        :raises ValueError: if something happens
    """

    def __init__(self, args: T.List[str], description: str) -> None:
        """
        Initialize self.

        :param args: list of arguments before delimiting colon.
        :param description: associated docstring description.
        """
        self.args = args
        self.description = description

    @classmethod
    def from_meta(cls, meta: "DocstringMeta") -> T.Any:
        """Copy DocstringMeta from another instance."""
        return cls(args=meta.args, description=meta.description)


class DocstringTypeMeta(DocstringMeta):
    """Docstring meta whose only optional arg contains type information."""

    @property
    def type_name(self) -> T.Optional[str]:
        """Return type name associated with given docstring metadata."""
        return self.args[1] if len(self.args) > 1 else None


class DocstringParam(DocstringMeta):
    """DocstringMeta symbolizing :param metadata."""

    @property
    def arg_name(self) -> T.Optional[str]:
        """Return argument name associated with given param."""
        if len(self.args) > 2:
            return self.args[2]
        elif len(self.args) > 1:
            return self.args[1]
        return None

    @property
    def type_name(self) -> T.Optional[str]:
        """Return type name associated with given param."""
        return self.args[1] if len(self.args) > 2 else None


class DocstringReturns(DocstringTypeMeta):
    """DocstringMeta symbolizing :returns metadata."""

    pass


class DocstringRaises(DocstringTypeMeta):
    """DocstringMeta symbolizing :raises metadata."""

    pass


class DocstringExamples(DocstringTypeMeta):
    """DocstringMeta symbolizing :examples metadata."""

    @property
    def name(self) -> T.Optional[str]:
        """Return the example name associated with given param."""
        if self.args:
            return self.args[1]
        return None


class Docstring:
    """Docstring object representation."""

    def __init__(self) -> None:
        """Initializes self."""
        self.short_description = None  # type: T.Optional[str]
        self.long_description = None  # type: T.Optional[str]
        self.blank_after_short_description = False
        self.blank_after_long_description = False
        self.meta = []  # type: T.List[DocstringMeta]

    @property
    def params(self) -> T.List[DocstringParam]:
        """Return parameters indicated in docstring."""
        return [
            DocstringParam.from_meta(meta)
            for meta in self.meta
            if meta.args[0]
            in {"param", "parameter", "arg", "argument", "key", "keyword"}
        ]

    @property
    def raises(self) -> T.List[DocstringRaises]:
        """Return exceptions indicated in docstring."""
        return [
            DocstringRaises.from_meta(meta)
            for meta in self.meta
            if meta.args[0] in {"raises", "raise", "except", "exception"}
        ]

    @property
    def returns(self) -> T.Optional[DocstringReturns]:
        """Return return information indicated in docstring."""
        try:
            return next(
                DocstringReturns.from_meta(meta)
                for meta in self.meta
                if meta.args[0] in {"return", "returns", "yield", "yields"}
            )
        except StopIteration:
            return None

    @property
    def examples(self) -> T.Optional[DocstringExamples]:
        """Return example information indicated in docstring."""
        return [
            DocstringExamples.from_meta(meta)
            for meta in self.meta
            if meta.args[0] in {"example", "examples"}
        ]



"""
Original © NetApp 2024

This is the main staring point for a recline application
"""

import atexit
import os
import readline
import shlex
import signal
import sys
import traceback
from typing import Callable, List, Union

import recline
from recline import commands
from recline.arg_types.recline_type_error import ReclineTypeError
from recline.commands import ReclineCommandError, builtin_commands
from recline.commands.async_command import AsyncCommand, CommandBackgrounded, CommandCancelled
from recline.repl import completer


# pylint: disable=too-many-arguments,bad-continuation
def relax(
    argv: str = None,
    program_name: str = None,
    motd: Union[Callable[[], str], str] = None,
    history_file: str = None,
    prompt: str = None,
    repl_mode: bool = True,
    single_command: str = None,
) -> None:
    """This is the main entry point of a recline-based application. Call this after
    all of your commands have been defined.

    This will start the application in its REPL mode (by default) or in a single
    command mode if the first argv to is set to '-c'. If in REPL mode, this function
    will not return until the user quits the application.

    Args:
        argv: This is a list of arguments that usually comes from the command line.
            This will default to sys.argv[1:] but may be overridden if needed.
        program_name: If provided, this will be the name of the program (used
            internally for display purposes). If not provided, sys.argv[0] will
            be used as a default (the name of the script file, e.g. hello.py).
        motd: If provided, this will be displayed to the user before the prompt.
            If it is a string, it will be displayed as given. If it is a callable,
            then the result will be displayed to the user.
        history_file: If persistent command history is desired, a file name may
            be passed where a list of the most recent commands will be kept and
            loaded from.
        prompt: This is the prompt to display to the user to let them know that
            the system is ready to accept a command. If not provided, a default
            will be used.
        repl_mode: By default, applications using recline expect to act as a REPL
            environment where the user will run many commands. These applications
            can still run a single command if the user passes a -c. But, it is
            convienent for some applications to act more like a traditional CLI
            command without needing to pass -c. For those applicatinos, repl_mode
            can be set to False.
        single_command: This acts a bit like non-repl mode, except in this case
            the user doesn't have to pass any command name as an argument to the
            application, but rather whatever command name is passed here is automatically
            run and the application exits when the command is complete. If an
            application only had one command, such as an implementation of ls in
            Python, then this would be a good choice for exposing that command easily.
    """

    if argv is None:
        argv = sys.argv[:]

    _setup_repl(program_name, prompt, history_file, argv)

    if single_command or not repl_mode or argv and len(argv) >= 2 and argv[1] == "-c":
        # Some external process sent us a command to run (like scp or ssh or
        # so run it and exit
        command = argv[2:]
        if not repl_mode:
            command = argv[1:]
        if single_command:
            command = [single_command] + argv[1:]
        return run_one_command(f"{' '.join(command)}")

    if isinstance(motd, str):
        print(motd)
    elif motd:
        print(motd())

    # The main loop (the L in REPL)
    while True:
        try:
            current_input = input(recline.PROMPT).strip()  # the R in REPL
            if current_input == "":
                continue
            execute(current_input)  # The E and P in REPL
        except (KeyboardInterrupt, CommandCancelled):
            # This is what bash looks like if you interrupt it. It will print out
            # the ^C character and go to the next line.
            sys.stdout.write("^C\n")
        except builtin_commands.DebugInterrupt:
            # This makes it look less confusing after returning from a debug break
            # otherwise the prompt isn't seen unless the user presses enter
            sys.stdout.write("\n")
            builtin_commands.debug()
        except EOFError:
            # If the terminal is closed on us (maybe SSH exit), just exit the app
            sys.exit(0)


def execute(current_input: str) -> int:
    """Execute the input as a series of commands"""

    result = 0

    if ';' in current_input:
        for command in current_input.split(';'):
            result = execute(command)
        return result

    if '&&' in current_input:
        for command in current_input.split('&&'):
            result = execute(command)
            if result > 0:
                break
        return result

    if "||" in current_input:
        for command in current_input.split('||'):
            result = execute(command)
            if result == 0:
                break
        return result

    # if it is a simple command, run it
    return run_one_command(current_input.strip())


def run_one_command(current_input: str) -> int:
    """ Parse the current input, search for a matching command and execute it
        with the given parameters.
    """

    command_name = completer.CommandCompleter.get_command_name(
        current_input, list(commands.COMMAND_REGISTRY.keys()),
    )
    if command_name not in commands.COMMAND_REGISTRY:
        print(f"Unknown command: {current_input}")
        return 1
    command = commands.COMMAND_REGISTRY[command_name]
    cmd_args = [f for f in shlex.split(current_input.replace(command_name, "", 1)) if f]
    if cmd_args and cmd_args[-1] == "?" and command_name != "?":
        cmd_args[-1] = "-help"
    try:
        result = _run_command(command, cmd_args)
        if command.output_formatter:
            command.output_formatter.format_output(result)
            # the command ran to completion without exception so we'll say the
            # return code is 0
            return 0
        return result
    except CommandBackgrounded as err:
        print(f'^Z\nJob {err.job_pid} is running in the background')
        return 0
    except CommandCancelled:
        pass
    except (ReclineTypeError, ReclineCommandError) as exc:
        print(str(exc))
    except Exception as exc:  # pylint: disable=broad-except
        print(f"Command execution error: {exc}")
        traceback.print_exc()
    except SystemExit as exc:
        if exc.code == builtin_commands.EXIT_COMMAND_CODE:
            raise

        # if a command fails to parse, that's OK. The user probably just didn't
        # type one of the required parameters yet. We might also get here if they
        # printed the help output with -help
        return exc.code
    return 1


def _setup_repl(program_name: str, prompt: str, history_file: str, argv: List[str]) -> None:
    if not program_name:
        recline.PROGRAM_NAME = argv[0].split(os.path.sep)[-1]
    else:
        recline.PROGRAM_NAME = program_name

    if prompt is not None:
        recline.PROMPT = prompt

    if history_file:
        track_command_history(history_file)
    setup_tab_complete()

    if sys.platform != "win32":
        # set up a handler for ctrl-\
        def signal_handler(signum, frame):
            raise builtin_commands.DebugInterrupt()
        signal.signal(signal.SIGQUIT, signal_handler)

    if commands.EXIT_COMMAND:
        atexit.register(_run_command, commands.EXIT_COMMAND, [])

    if commands.START_COMMAND:
        try:
            _run_command(commands.START_COMMAND, argv[1:])
        except CommandBackgrounded:
            pass


def _run_command(command: str, cmd_args: List[str]) -> int:
    namespace = command.parser.parse_args(args=cmd_args)

    args = [getattr(namespace, arg.name) for arg in command.required_args]
    kwargs = {
        arg.name: getattr(namespace, arg.name)
        for arg in command.optional_args if hasattr(namespace, arg.name)
    }

    result = None
    if command.is_async:
        command_thread = AsyncCommand(command, *args, **kwargs)
        command_thread.start()
        if namespace.background or command.is_background:
            raise CommandBackgrounded(command_thread.job_pid)
        result = command_thread.foreground()
    else:
        result = command.func(*args, **kwargs)
    if result is None:
        result = 0
    return result


def track_command_history(filename: str) -> None:
    """load/save command history"""

    try:
        readline.read_history_file(filename)
        # default history len is -1 (infinite), which may grow unruly
        readline.set_history_length(1000)
    except IOError:
        pass
    atexit.register(readline.write_history_file, filename)


def setup_tab_complete() -> None:
    """Set up the readline library to hook into our command completer"""

    readline.set_completer_delims("")
    if sys.platform == "linux":
        # pyreadline3 (used for windows) doesn't implement this
        # libedit (default for macos) implements this but never calls it (broken)
        readline.set_completion_display_matches_hook(completer.match_command_hook)
    readline.set_completer(completer.CommandCompleter(commands.COMMAND_REGISTRY).completer)
    if readline.__doc__ and "libedit" in readline.__doc__:
        # macos doesn't use GNU readline, but BSD libedit instead
        readline.parse_and_bind("bind '\t' rl_complete")
        readline.parse_and_bind("bind '^R' em-inc-search-prev")
    else:
        readline.parse_and_bind("tab: complete")
    readline.parse_and_bind("set show-all-if-ambiguous on")
    readline.parse_and_bind("set bell-style none")



"""
Original © NetApp 2024

A Flag type allows the CLI command writer to specify an argument that behaves
like a boolean flag. The difference between this and declaring an argument as a
bool is that no "true" or "false" input is expected of the user if the argument
is a flag. If it is a bool, then the user must provide either "true" or "false"
explicitly on the command line.

@recline.command
def ls(l: Flag = None) -> None:
    # If the user provided "-l" then l will be set to True. Else l will be None
"""

from recline.arg_types.recline_type import ReclineType


class Flag(ReclineType):
    """A simple flag argument which will be set to true if provided"""
    action = 'store_true'



"""
Original © NetApp 2024

This module contains the abstract class that all other output formatters should
inherit from
"""

from abc import ABC, abstractmethod

class OutputFormatter(ABC):  # pylint: disable=too-few-public-methods
    """All output formatters should inherit from this class"""

    @abstractmethod
    def format_output(self, results):
        """The implementation of this method should expect to receive the results
        of a command output and to process them and output them in whatever manner
        is appropriate for the specific formatter.
        """



"""
Original © NetApp 2024

This module contains a table formatter for CLI output. The table formatter expects
output to be an iterable of dictionary-like objects. For example, if a CLI command
returned something like:

[{'foo': 'bar', 'count': 1}, {'foo': 'baz', 'count': 2}]

Then the table formatter would print this to the screen:

+-----+-------+
| Foo | Count |
+-----+-------+
| bar | 1     |
| baz | 2     |
+-----+-------+
"""

from collections import OrderedDict

from recline.formatters.output_formatter import OutputFormatter


class TableFormat(OutputFormatter):  # pylint: disable=too-few-public-methods
    """A TableFormat object can be used as the return type for a CLI command. When
    specified, the returned value from the CLI command is assumed to be an interable
    of dictionary-like objects. All of the objects are assumed to be homogeneous
    with respsect to their keys. It will print out a table representing the data.
    """

    def format_output(self, results):
        """If there are no results given to us, print a message, otherwise print
        a table with a row for each record in the results.
        """

        if not results:
            print('No records')
            return

        columns = list(results[0].keys())
        data = []
        for result in results:
            data.append(list(result.values()))
        self._print_table(columns, data)

    def _print_table(self, headers, data):  # pylint: disable=no-self-use
        """Iterate the data and print lines for the headers and the cells"""

        col_data = OrderedDict()
        row_data = []
        col_widths = [0] * len(headers)
        heading_index = 0
        for heading in headers:
            # longest piece of data for each column
            longest = len(heading)
            for row in data:
                if len(str(row[heading_index])) > longest:
                    longest = len(str(row[heading_index]))

            col_widths[heading_index] = longest
            heading_index += 1
            col_data[heading_index] = (heading, longest)

        for row in data:
            row_dict = OrderedDict()
            heading_index = 0
            for cell in row:
                row_dict[heading_index] = (cell, col_widths[heading_index])
                heading_index += 1
            row_data.append(row_dict)

        _print_border(col_data)
        _print_data_line(col_data, headers=True)
        _print_border(col_data)

        for row in row_data:
            _print_data_line(row)

        _print_border(col_data)


def _print_border(col_data):
    """Print a top or bottom border on the table depending on its width"""

    for _, width in col_data.values():
        print('+', end='')
        print('-' * (width + 2), end='')
    print('+')


def _print_data_line(col_data, headers=False):
    """Print a line of data including the cell borders"""

    for _, (data, width) in col_data.items():
        if headers:
            data = str(data).title().replace('_', ' ')
        print(f'| {data}', end='')
        print(' ' * (width - len(str(data)) + 1), end='')
    print('|')



"""
Original © NetApp 2024

pytest configuration for recline unit tests
"""

import pytest

import recline


@pytest.fixture(scope='function')
def clean_jobs():
    """Make sure the globals are fresh in each test"""

    recline.JOBS = {}
    recline.NEXT_JOB_PID = 1
    recline.commands.COMMAND_REGISTRY = {}



"""
Original © NetApp 2024

A Positional type allows the CLI command writer to specify an argument that the
user will provide positionally. That is, instead of the user typing
"-arg_name argvalue", they can just type "argvalue". For example:

@recline.command
def ls(path: Positional = ".") -> None:
    # If the user provided "ls my_dir", then path will be set to "my_dir". If the
    # user just said "ls", then path will be set to ".".
"""

from recline.arg_types.recline_type import ReclineType
from recline.arg_types.recline_type_error import ReclineTypeError


class Positional(ReclineType):
    """The Positional type allows the user to not have to specify the argument name"""

    @staticmethod
    def define(completer=None, data_type=str):
        """Allows a Positional type to be defined with its own completer function"""

        class _Positional(Positional):
            def __init__(self):
                if completer:
                    self.completer = completer
                self.data_type = data_type
                super().__init__()

            def validate(self, arg: str):
                if arg is None or arg == "":
                    raise ReclineTypeError("Value cannot be empty")
                try:
                    return self.data_type(arg)
                except Exception:
                    raise ReclineTypeError(
                        f"Cannot parse {arg} as a valid {data_type.__name__}"
                    )

        return _Positional



"""ReST-style docstring parsing."""

import inspect
import re

from .common import Docstring, DocstringMeta, ParseError


def parse(text: str) -> Docstring:
    """
    Parse the ReST-style docstring into its components.

    :returns: parsed docstring
    """
    ret = Docstring()
    if not text:
        return ret

    text = inspect.cleandoc(text)
    match = re.search("^:", text, flags=re.M)
    if match:
        desc_chunk = text[: match.start()]
        meta_chunk = text[match.start() :]
    else:
        desc_chunk = text
        meta_chunk = ""

    parts = desc_chunk.split("\n", 1)
    ret.short_description = parts[0] or None
    if len(parts) > 1:
        long_desc_chunk = parts[1] or ""
        ret.blank_after_short_description = long_desc_chunk.startswith("\n")
        ret.blank_after_long_description = long_desc_chunk.endswith("\n\n")
        ret.long_description = long_desc_chunk.strip() or None

    for match in re.finditer(
        r"(^:.*?)(?=^:|\Z)", meta_chunk, flags=re.S | re.M
    ):
        chunk = match.group(0)
        if not chunk:
            continue
        try:
            args_chunk, desc_chunk = chunk.lstrip(":").split(":", 1)
        except ValueError:
            raise ParseError('Error parsing meta information near "%s".' % chunk)
        args = args_chunk.split()
        desc = desc_chunk.strip()
        if "\n" in desc:
            first_line, rest = desc.split("\n", 1)
            desc = first_line + "\n" + inspect.cleandoc(rest)
        ret.meta.append(DocstringMeta(args, description=desc))

    return ret



"""
Original © NetApp 2024

This package contains useful output formatters. An output formatter can be defined
as the return type of a CLI command (using a type annotation). If one is defined,
then the result of running the command will be passed to the output formatter
which will usually print the result to the screen for the user.
"""



"""
Original © NetApp 2024

A Choices type allows the CLI command writer to specify a static list of choices
for a parameter. Once the body of the function is invoked, it is guaranteed that
the validation was done on the parameter to make sure it matched one.

@recline.command(name="cake make")
def make_cake(flavor: Choices.define(["chocolate", "vanilla", "marble"])) -> None:
    # We can assume flavor is one of the choices in the body of the function
"""

from typing import Callable, List, Union

from recline.arg_types.recline_type import ReclineType
from recline.arg_types.recline_type_error import ReclineTypeError


class Choices(ReclineType):
    """The Choices type validates the user input according to a list of possible choices"""

    @staticmethod
    def define(
            available_choices: Union[List, Callable], cache_choices: bool = False,
            inexact: bool = False, data_type=str,
        ) -> "Choices":
        """A `recline.commands.types.Choices` is a way to assert that an argument
        is one of a predefined list.

        Prior to being passed to your command, the argument will be compared to
        the list of defined choices. If it does not match one, an error will be
        printed.

        Args:
            available_choices: The list of allowable values for the argument or
                a callable that will return a list.
            cache_choices: If the available_choices argument was a callable, by
                default it will be called each time the list of choices needs to
                be rendered. But if that callable is expensive, that may not be
                desirable. In that case, passing cache_choices as True will call
                the available_choices function only once to get the list and save
                the result.
            inexact: If set to True, allow the value being validated to contain
                the "*", "|", or ".." query characters. If these characters are
                present, then no validation will be done on the value (and validation
                is assumed to be done on whatever the command is calling)
            data_type: The type of data that represents this argument
        """

        class _Choices(Choices):
            def __new__(cls):
                instance = super().__new__(cls)
                cls.available_choices = available_choices
                if isinstance(available_choices, list):
                    cls.metavar = f"<{'|'.join(available_choices)}>"
                return instance

            def validate(self, arg):
                if inexact:
                    return arg
                current_choices = self.choices(eager=True)
                if arg not in current_choices:
                    raise ReclineTypeError(
                        f"\"{arg}\" must be one of {', '.join(current_choices)}."
                    )
                try:
                    return data_type(arg)
                except Exception:  # pylint: disable=broad-except
                    raise ReclineTypeError(f'Unable to convert "{arg}" to type {data_type}')

            def choices(self, eager=False):
                if hasattr(self.__class__, '_cached_choices'):
                    return self.__class__._cached_choices  # pylint: disable=protected-access

                # don't call the completer function unless we're sure we want to
                if not eager and not isinstance(available_choices, list):
                    return None

                current_choices = self.__class__.available_choices
                try:
                    current_choices = self.__class__.available_choices()
                except TypeError:
                    pass

                current_choices = [str(c) for c in current_choices]
                if cache_choices:
                    self.__class__._cached_choices = current_choices  # pylint: disable=protected-access
                self.__class__.metavar = f"<{'|'.join(current_choices)}>"
                return current_choices

            def completer(self, *args, **kwargs):
                return self.choices(eager=True)

        return _Choices



"""Docstring parsing."""

import enum

from . import google, rest
from .common import Docstring, ParseError


class Style(enum.Enum):
    rest = 1
    google = 2
    auto = 3


_styles = {Style.rest: rest.parse, Style.google: google.parse}


def parse(text: str, style: Style = Style.auto) -> Docstring:
    """
    Parse the docstring into its components.

    :param text: docstring text to parse
    :param style: docstring style
    :returns: parsed docstring representation
    """

    if style != Style.auto:
        return _styles[style](text)
    rets = []
    for parse_ in _styles.values():
        try:
            rets.append(parse_(text))
        except ParseError as e:
            exc = e
    if not rets:
        raise exc
    return sorted(rets, key=lambda d: len(d.meta), reverse=True)[0]



"""Google-style docstring parsing."""

import inspect
import re

from .common import Docstring, DocstringMeta, ParseError

_sections = {
    "Arguments": "param",
    "Args": "param",
    "Parameters": "param",
    "Params": "param",
    "Raises": "raises",
    "Exceptions": "raises",
    "Except": "raises",
    "Attributes": None,
    "Example": "example",
    "Examples": "example",
    "Returns": "returns",
    "Yields": "yields",
}
_titles_re = re.compile(
    "^(" + "|".join("(%s)" % t for t in _sections) + "):", flags=re.M
)
_valid = {t for t, a in _sections.items() if a}


def _build_meta(text: str, title: str) -> DocstringMeta:
    """Build docstring element.

    :param text: docstring element text
    :param title: title of section containing element
    :return:
    """

    meta = _sections[title]
    if meta == "returns" and ":" not in text.split()[0]:
        return DocstringMeta([meta], description=text)

    # Split spec and description
    before, desc = text.split(":", 1)
    if desc:
        desc = desc[1:] if desc[0] == " " else desc
        if "\n" in desc:
            first_line, rest = desc.split("\n", 1)
            desc = first_line + "\n" + inspect.cleandoc(rest)
        desc = desc.strip("\n")

    # Build Meta args
    m = re.match(r"(\S+) \((\S+)\)$", before)
    if meta == "param" and m:
        arg_name, type_name = m.group(1, 2)
        args = [meta, type_name, arg_name]
    else:
        args = [meta, before]

    return DocstringMeta(args, description=desc)


def parse(text: str) -> Docstring:
    """
    Parse the Google-style docstring into its components.

    :returns: parsed docstring
    """
    ret = Docstring()
    if not text:
        return ret

    # Clean according to PEP-0257
    text = inspect.cleandoc(text)

    # Find first title and split on its position
    match = _titles_re.search(text)
    if match:
        desc_chunk = text[: match.start()]
        meta_chunk = text[match.start() :]
    else:
        desc_chunk = text
        meta_chunk = ""

    # Break description into short and long parts
    parts = desc_chunk.split("\n", 1)
    ret.short_description = parts[0] or None
    if len(parts) > 1:
        long_desc_chunk = parts[1] or ""
        ret.blank_after_short_description = long_desc_chunk.startswith("\n")
        ret.blank_after_long_description = long_desc_chunk.endswith("\n\n")
        ret.long_description = long_desc_chunk.strip() or None

    # Split by sections determined by titles
    matches = list(_titles_re.finditer(meta_chunk))
    if not matches:
        return ret
    splits = []
    for j in range(len(matches) - 1):
        splits.append((matches[j].end(), matches[j + 1].start()))
    splits.append((matches[-1].end(), len(meta_chunk)))

    chunks = {}
    for j, (start, end) in enumerate(splits):
        title = matches[j].group(1)
        if title not in _valid:
            continue
        chunks[title] = meta_chunk[start:end].strip("\n")
    if not chunks:
        return ret

    # Add elements from each chunk
    for title, chunk in chunks.items():
        # Determine indent
        indent_match = re.search(r"^\s+", chunk)
        if not indent_match:
            raise ParseError('Can\'t infer indent from "%s"' % chunk)
        indent = indent_match.group()

        # Check for returns/yields (only one element)
        if _sections[title] in ("returns", "yields"):
            part = inspect.cleandoc(chunk)
            ret.meta.append(_build_meta(part, title))
            continue

        # Split based on lines which have exactly that indent
        _re = "^" + indent + r"(?=\S)"
        c_matches = list(re.finditer(_re, chunk, flags=re.M))
        if not c_matches:
            raise ParseError('No specification for "%s": "%s"' % (title, chunk))
        c_splits = []
        for j in range(len(c_matches) - 1):
            c_splits.append((c_matches[j].end(), c_matches[j + 1].start()))
        c_splits.append((c_matches[-1].end(), len(chunk)))
        for j, (start, end) in enumerate(c_splits):
            part = chunk[start:end].strip("\n")
            ret.meta.append(_build_meta(part, title))

    return ret



"""
Original © NetApp 2024

A test module for the recline.formatters module
"""

from collections import OrderedDict

import pytest

from recline.formatters.table_formatter import TableFormat


@pytest.mark.parametrize('command_results,expected_table', [
    ([], "No records\n"),
    (
        [OrderedDict([('foo', 'bar'), ('baz', 'bleep')])],
        "+-----+-------+\n"
        "| Foo | Baz   |\n"
        "+-----+-------+\n"
        "| bar | bleep |\n"
        "+-----+-------+\n",
    ),
    (
        [
            OrderedDict([
                ('foo_su', 'bar'), ('baz', 'bleep'),
            ]),
            OrderedDict([
                ('foo_su', 'bar2'), ('baz', 'bleep2'),
            ]),
        ],
        "+--------+--------+\n"
        "| Foo Su | Baz    |\n"
        "+--------+--------+\n"
        "| bar    | bleep  |\n"
        "| bar2   | bleep2 |\n"
        "+--------+--------+\n",
    ),
])
def test_table_format(command_results, expected_table, capsys):
    """Verify that our formatter gives the results we expect depending on the input"""

    TableFormat().format_output(command_results)
    captured = capsys.readouterr()
    assert captured.out == expected_table



"""
Original © NetApp 2024

This module contains the error class that is raised when there is a type error
when validating user input.
"""

import inspect

from recline.arg_types.recline_type import ReclineType


class ReclineTypeError(Exception):
    """This is a custom exception type that tries to reach into the previous frame
    and get some information if it is present in order to make its error message
    better.
    """

    def __init__(self, message: str):
        # Reach up frame and see if we can attach the arg name to the message
        # If we can't for some reason, it won't get attached
        try:
            type_instance = inspect.currentframe().f_back.f_locals.get("self")
            if isinstance(type_instance, ReclineType) and type_instance.arg_name is not None:
                super().__init__(f"{type_instance.arg_name}: {message}")
        except Exception:  # pylint: disable=broad-except
            super().__init__(message)



"""
Original © NetApp 2024

This module contains the definition of a CLICommand object. This is the main
driver of the recline library. Essentially, a CLICommand takes as input a wrapped
function and parses the args and docstring so that when called from the REPL, the
command can be used as prescribed or the help information can be queried.
"""

import argparse
from functools import wraps, partial
import inspect
import io
import json
import logging
from typing import Callable, List, Union

from recline import commands
from recline.arg_types.recline_type import ReclineType, UniqueParam
from recline.arg_types.positional import Positional
from recline.arg_types.flag import Flag
from recline.arg_types.remainder import Remainder
from recline.formatters.output_formatter import OutputFormatter
from recline.vendor import docstring_parser


LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())


class CLICommand:  # pylint: disable=too-many-instance-attributes
    """This is the implementation of a @recline.command. It sets up all of the
    parameter validation, help text generation, and return output formatting for
    commands defined in the application.
    """

    # pylint: disable=too-many-arguments,bad-continuation
    def __init__(
        self, func,
        name=None,
        group=None,
        is_alias=False,
        hidden=False,
        is_async=False,
        is_background=False,
    ):
        self.func = func
        self.name = name if name else func.__name__
        self.is_alias = is_alias
        self.group = group
        self.is_async = is_async
        self._hidden = hidden
        self.is_background = is_background

        signature = inspect.signature(func)
        parameters = signature.parameters
        self.required_args = [
            p for p in parameters.values()
            if p.default == inspect.Parameter.empty
        ]
        self.optional_args = [
            p for p in parameters.values()
            if p.default != inspect.Parameter.empty
        ]
        self.docstring = docstring_parser.parse(self.func.__doc__)

        return_type = signature.return_annotation
        if return_type and issubclass(return_type, OutputFormatter):
            self.output_formatter = return_type()
        else:
            self.output_formatter = None
        self.parser = self._create_parser()

        # Using a separate parser object to pass to argcomplete prevents an issue
        # with custom types and custom validate methods. Haven't been able to nail
        # down the exact cause though.
        self.completer_parser = self._create_parser()

    def __str__(self):
        return (
            '%s(%s, *, %s)' %
            (
                self.name, ', '.join([p.name for p in self.required_args]),
                ', '.join(['%s=%s' % (p.name, p.default) for p in self.optional_args]),
            )
        )

    @property
    def hidden(self):
        """A command is hidden if it is either statically declared to be or it
        was provided with a function to determine if it should currently be hidden
        """

        if isinstance(self._hidden, bool):
            return self._hidden
        return self._hidden()

    def _create_parser(self):
        parser = argparse.ArgumentParser(
            add_help=False,
            description=(
                '%s\n\n%s' %
                (self.docstring.short_description, self.docstring.long_description)
            )
        )
        parser.add_argument('-help', help=argparse.SUPPRESS, action=self.print_help())

        if self.is_async:
            parser.add_argument(
                '-background', help="Start this command in the background",
                action="store_true",
            )

        arg_specs = []
        for arg in self.required_args:
            arg_specs.append(self._get_arg_spec(arg, True))
        for arg in self.optional_args:
            arg_specs.append(self._get_arg_spec(arg, False))

        for spec in arg_specs:
            arg = spec[2]
            action = parser.add_argument(*spec[0], **spec[1])
            action.completer = lambda *x, **y: [None]
            annotation_type = get_annotation_type(arg)
            if issubclass(annotation_type, ReclineType):
                if inspect.isclass(arg.annotation):
                    type_instance = arg.annotation()
                action.completer = type_instance.completer
            elif getattr(action, 'choices', None) is not None:
                choices = action.choices[:]
                action.completer = lambda *x, **y: choices  # pylint: disable=cell-var-from-loop

        return parser

    def _get_arg_spec(self, arg, required):
        spec_args = [f'-{arg.name}']
        spec_kwargs = {
            'action': self.get_arg_action(arg),
            'help': self.get_arg_description(arg),
            'metavar': self.get_arg_metavar(arg),
            'required': required,
        }

        if arg.default != inspect.Parameter.empty:
            spec_kwargs['default'] = arg.default

        def _recline_type_annotation_handler(arg_annotation, nargs=None):
            nonlocal spec_args, spec_kwargs, arg

            if inspect.isclass(arg_annotation):
                type_instance = arg_annotation()
            type_instance.arg_name = arg.name
            spec_kwargs['type'] = type_instance.validate
            choices = type_instance.choices()
            if choices is not None:
                spec_kwargs['choices'] = choices
            spec_kwargs['nargs'] = nargs or type_instance.nargs()

            special_nargs = (
                spec_kwargs['nargs'] == argparse.REMAINDER or
                isinstance(type_instance, Positional)
            )
            if special_nargs:
                del spec_kwargs['required']
                spec_args = [arg.name]
                if isinstance(type_instance, Positional) and arg.default is not None:
                    spec_kwargs['nargs'] = '?'
            if isinstance(type_instance, Flag):
                del spec_kwargs['metavar']
                del spec_kwargs['type']
                del spec_kwargs['nargs']

        if arg.annotation:
            annotation_type = get_annotation_type(arg)
            if issubclass(annotation_type, List):
                if issubclass(arg.annotation.__args__[0], ReclineType):
                    _recline_type_annotation_handler(arg.annotation.__args__[0], '+')
                else:
                    spec_kwargs['nargs'] = '+'
            if issubclass(annotation_type, bool):
                spec_kwargs['choices'] = [True, False]
                spec_kwargs['type'] = lambda val: val.lower() == 'true'
            if issubclass(annotation_type, int) and not issubclass(annotation_type, bool):
                spec_kwargs['type'] = int
            if issubclass(annotation_type, float):
                spec_kwargs['type'] = float
            if issubclass(annotation_type, dict):
                spec_kwargs['type'] = json.loads
            if issubclass(annotation_type, ReclineType):
                _recline_type_annotation_handler(annotation_type)

        return spec_args, spec_kwargs, arg

    def get_arg_action(self, arg):  #pylint: disable=no-self-use
        """By default, arguments will be verified as unique. However, if a custom
        type defines an action that should be taken when parsing the command line,
        that action will be used instead.
        """

        action = UniqueParam
        if not arg.annotation:
            return action

        annotation_type = get_annotation_type(arg)
        if issubclass(annotation_type, ReclineType):
            return annotation_type.action
        return action

    def get_arg_metavar(self, arg):  #pylint: disable=no-self-use,too-many-return-statements
        """By default, the metavar for an argument is the same as the name of that
        argument (and enclosed in <>). But there are many situations with custom
        types where the metavar will be defined by the type and should be taken
        from the type.
        """

        basic = f'<{arg.name}>'
        if not arg.annotation:
            return basic

        annotation_type = get_annotation_type(arg)

        if issubclass(annotation_type, Positional):
            return basic
        if issubclass(annotation_type, ReclineType):
            return arg.annotation.metavar
        if issubclass(annotation_type, List):
            return f'<{arg.name}> [{arg.name} ...]'
        if issubclass(annotation_type, bool):
            return '<true|false>'
        if issubclass(annotation_type, int):
            return '<int>'
        return basic

    def get_arg_description(self, arg, indent=0):
        """Get the description of the arg from the parsed docstring if we can find
        one. Otherwise, the description will be blank.
        """

        for doc_param in self.docstring.params:
            if doc_param.arg_name == arg.name:
                if indent is None:
                    return doc_param.description.replace('\n', ' ')
                return doc_param.description.replace('\n', '\n' + ' ' * indent)
        return ''

    def get_arg_help(self, arg, indent=True):
        """Get a line of help text for an argument including the name, metavar,
        and description of the argument.
        """

        annotation_type = get_annotation_type(arg)
        positional = '' if issubclass(annotation_type, (Positional, Remainder)) else '-'
        name_meta = f'  {positional}{arg.name} {self.get_arg_metavar(arg)} '
        if issubclass(annotation_type, Positional):
            name_meta = f'  {self.get_arg_metavar(arg)} '
        if issubclass(annotation_type, Flag):
            name_meta = f'  {positional}{arg.name} '
        indent = len(name_meta) if indent else None
        arg_help = name_meta
        arg_description = self.get_arg_description(arg, indent=indent)
        if arg_description:
            arg_help += arg_description
        if arg.default != inspect.Parameter.empty and not issubclass(annotation_type, Flag):
            default_val = arg.default
            if isinstance(arg.default, bool):
                default_val = str(arg.default).lower()
            arg_help += f'\n    Default: {default_val}'
        return arg_help

    def get_command_help(self):
        """Get a block of help text for the command including all of its arguments
        as well as a short description of the command.
        """

        help_text = []
        if self.docstring.short_description:
            help_text.append(self.docstring.short_description)
        if self.required_args or self.optional_args:
            help_text.append('')
        if self.required_args:
            help_text.append('Required arguments:')
            for arg in self.required_args:
                help_text.append(self.get_arg_help(arg))
            if self.optional_args:
                help_text.append('')
        if self.optional_args:
            help_text.append('Optional arguments:')
            for arg in self.optional_args:
                help_text.append(self.get_arg_help(arg))

        return '\n'.join(help_text)

    def get_command_usage(self):
        """Get the short usage spec for the command"""

        str_file = io.StringIO()
        self.parser.print_usage(file=str_file)
        usage = str_file.getvalue()
        # argparse gives us a few pieces we need to replace
        usage = ' '.join(usage.split(' ')[2:])
        return self.name + ' ' + usage.strip()

    def print_help(outer_self):  # pylint: disable=no-self-argument
        """If the -help argument is passed to the command, then this method will
        print out the help for that command instead of the builtin argparse help printer.
        """

        class HelpAction(argparse.Action):  # pylint: disable=too-few-public-methods
            """This is a custom help action which implements the argparse action
            protocol.
            """

            # pylint: disable=bad-continuation
            def __init__(
                self,
                option_strings,
                dest=argparse.SUPPRESS,
                default=argparse.SUPPRESS,
                help=None  # pylint: disable=redefined-builtin
            ):
                super().__init__(option_strings, dest=dest, default=default, nargs=0, help=help)
            def __call__(self, parser, namespace, values, option_string=None):
                print(outer_self.get_command_help())
                parser.exit()
        return HelpAction


# pylint: disable=too-many-arguments,bad-continuation
def command(
    func=None,
    name: str = None,
    group: str = None,
    aliases: List[str] = None,
    atstart: bool = False,
    atexit: bool = False,
    hidden: Union[Callable[[], bool], bool] = False,
    background: bool = False
):
    """Wrapping a function with this registers it with the recline library and
    exposes it as a command the user of the application can call.

    The name of the command will match the name of the function that is being
    wrapped. The arguments of the command will match the arguments that the
    function accepts. Any arguments that are required by the function will be
    required arguments for the command. Any arguments that are optional (listed
    as keyword arguments) will be optional for the command.

    Args:
        name: If the command name shouldn't inherit the function name, it can be
            defined here. This can be especially useful if the command name has
            spaces in it.
        group: If certain commands should be logically grouped together, they
            should share the same group name.
        aliases: An optional list of other names that will map back to the same
            function. For example, if the function was called 'exit', a list of
            aliases might be ['quit', 'q'].
        atstart: If set to True, this command will run before the main REPL is
            available for processing other commands. If this command fails, the
            REPL won't become available.
        atexit: Before exiting the program, this command will be run and can be
            used to clean up any resources. This command should always take 0
            arguments.
        hidden: If the command is hidden, then it will not be shown in the help
            output or available for autocompletion. It will still be executable
            by the user if they type it out. This can be either a function which
            evaluates to True or False, or just a constant True or False.
        background: If the command is async and long running, then setting
            background to True may be advisable so that other commands can be
            run in the foreground in the meantime.
    """


    if func is None:
        return partial(
            command, name=name, group=group, aliases=aliases, atstart=atstart,
            atexit=atexit, hidden=hidden, background=background,
        )

    is_async = inspect.iscoroutinefunction(func)

    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    _register(wrapper, name, group, aliases, atstart, atexit, hidden, is_async, background)
    return wrapper


# pylint: disable=too-many-arguments
def _register(func, name, group, aliases, atstart, atexit, hidden, is_async, is_background):
    if atstart:
        if commands.START_COMMAND:
            raise RuntimeError(f'A start command is already defined: {func}')
        commands.START_COMMAND = CLICommand(func, is_async=is_async, is_background=is_background)
        return
    if atexit:
        if commands.EXIT_COMMAND:
            raise RuntimeError(f'An exit command is already defined: {func}')
        commands.EXIT_COMMAND = CLICommand(func)
        return

    command_name = func.__name__
    if name is not None:
        command_name = name

    if not aliases:
        aliases = [command_name]
    else:
        aliases.insert(0, command_name)

    for index, alias in enumerate(aliases):
        is_alias = index > 0
        commands.COMMAND_REGISTRY[alias] = CLICommand(
            func, name=alias, group=group, is_alias=is_alias, hidden=hidden,
            is_async=is_async,
        )
        if not is_alias:
            LOGGER.debug('Registered %s', commands.COMMAND_REGISTRY[alias])
        else:
            LOGGER.debug('Aliased %s to %s', alias, command_name)


def get_annotation_type(arg):
    """Python 3.7 explicitly broke the issubclass checking on the typing module
    https://bugs.python.org/issue34568
    """

    return getattr(arg.annotation, '__origin__', arg.annotation)



"""
Original © NetApp 2024

This module implements a readline completer for completing command names as well
as argument names and values.
"""

import argcomplete
import argparse
import re
import readline
import sys

import recline
from recline.arg_types.recline_type import UniqueParam


def match_command_hook(substitution, matches, *_):
    """ Print a list of completion choices in columns and reprint the prompt """

    end = readline.get_endidx()
    current_line = readline.get_line_buffer()

    # back up the substitution string to the last full word in case of
    # partial matching
    substitution = ' '.join(substitution.split(' ')[:-1])
    completions = []
    for match in matches:
        subbed_match = match.replace(substitution, '').strip()
        if subbed_match:
            completions.append(subbed_match)

    sys.stdout.write(
        '\n%s\n%s%s\r%s%s' % (
            print_columns(completions), recline.PROMPT, current_line,
            recline.PROMPT, current_line[:end],
        )
    )


def print_columns(objects, cols=3, gap=2):
    """ Print a list of items in rows and columns """

    # if the list is empty, do nothing
    if not objects:
        return ""

    # make sure we have a string for each item in the list
    str_list = [str(x) for x in objects]

    # can't have more columns than items
    if cols > len(str_list):
        cols = len(str_list)

    max_length = max([len(x) for x in str_list])

    # get a list of lists which each represent a row in the output
    row_list = [str_list[i:i+cols] for i in range(0, len(str_list), cols)]

    # join together each row in the output with a newline
    # left justify each item in the list according to the longest item
    output = '\n'.join([
        ''.join([x.ljust(max_length + gap) for x in row_item])
        for row_item in row_list
    ])

    return output


class CommandCompleter:
    """A completer object that can be used to get the possible next choices for
    the user's current input. Used with readline's set_completer() function.
    """

    def __init__(self, commands):
        self.options = sorted(commands.keys())
        self.commands = commands
        self.matches = self.options[:]

    @staticmethod
    def get_command_name(text, options, exact=True):
        """ The text contains a complete command if there is 1 and only 1 command
            which contains the text's prefix
        """

        if exact and text in options:
            return text

        pieces = text.split(' ')
        i = len(pieces)
        command_name = None
        possible_match = None
        while i > 0:
            prefix = ' '.join(pieces[:i])
            possibles = [x for x in options if x.startswith(prefix)]
            if len(possibles) == 1:
                command_name = prefix
                possible_match = possibles[0]
                break
            if len(possibles) > 1:
                # This block covers the case where there are two commands with
                # the same prefix and the user has started typing args
                # Something like this:
                # possibles = ["deploy", "deploy status"]
                # text = "deploy 5"

                # walk each word of each possible
                # for each word, compare it to each word of the text
                # if the possible word starts with the matching text word, then
                # it is still a possible. Otherwise, it's not possible anymore
                still_possible = []
                for possible in possibles:
                    for index, possible_word in enumerate(possible.split(' ')):
                        if index > len(pieces) - 1:
                            continue
                        if not possible_word == pieces[index]:
                            break
                    else:
                        still_possible.append(possible)
                if len(still_possible) == 1:
                    command_name = still_possible[0]
                    possible_match = still_possible[0]
                    break

            i -= 1

        if command_name is not None and command_name == possible_match.strip():
            return command_name
        return None

    def _get_command_matches(self, command_str, text):
        """ Get the possible matches from a command by calling its completer """

        parser = self.commands[command_str].completer_parser
        completer = argcomplete.CompletionFinder(
            parser, always_complete_options='long'
        )
        matches = []
        state = 0
        for match in iter(lambda: completer.rl_complete(text, state), None):
            if not _already_added_match(parser, match, text):
                matches.append(match)
            state += 1

        return sorted(matches)

    def _parse_around_booleans(self, text):  # pylint: disable=no-self-use
        """In order to get completions for commands which might be part of a
        complex command line, we need to discard the parts of the complex command
        that we're not currently working on.

        For example, if the user was working on this sort of statement:

            my command -arg 1 && my other command -type ┋ ; my third command

        The user's cursor is where you see the ┋ and they have pressed tab. We
        need to know to only consider the "my other command -type " part of the
        line and nothing else
        """

        # when readline gives us the text, it only gives us the text up until
        # our cursor so we don't need to worry about anything after it
        right_hand_side = text.split("&&")[-1].split("||")[-1].split(";")[-1]
        left_hand_side = text[::-1].replace(right_hand_side[::-1], '', 1)[::-1]
        return left_hand_side, right_hand_side.lstrip()

    def completer(self, text, state):
        """To be used with readline's set_completer() function"""

        response = None
        prefix, text = self._parse_around_booleans(text)
        if state == 0:
            # if the text on the line already contains a complete command, we'll
            # need to get completions from that command instead
            command_name = self.get_command_name(text, self.options, exact=False)
            if command_name:
                self.matches = self._get_command_matches(command_name, text)
                if not self.matches and command_name == text:
                    self.matches.append(command_name + ' ')
            # This is the first time for this text, so build a match list
            elif text:
                self.matches = [x for x in self.options if x.startswith(text.strip())]
            else:
                self.matches = self.options[:]

        # Return the state'th item from the match list, if we have that many
        try:
            response = self.matches[state]
        except Exception:  # pylint: disable=broad-except
            return None

        current_line = readline.get_line_buffer()
        response = prefix + ' ' + response.rstrip()
        if not current_line.replace(prefix, '').lstrip().replace(text, '').startswith(' '):
            response += ' '
        return response.lstrip()


def _already_added_match(parser, match, text):
    """ Return True if the match given is not already specified in the text
        or if it is not a unique action and can be specified multiple times.
    """

    # pylint: disable=protected-access

    match = match.strip().split(' ')[-1]
    match_already_added = False
    unique_classes = [UniqueParam, argparse._StoreTrueAction, argparse._StoreFalseAction]

    for action in parser._actions:
        if match not in action.option_strings:
            continue

        action_is_unique = action._orig_class in unique_classes
        action_is_present = re.search(f' {match}', text)

        if action_is_unique and action_is_present:
            match_already_added = True
            break

    return match_already_added



"""
Original © NetApp 2024

A test module for the recline.commands.cli_command module
"""

from typing import List

import pytest

import recline
from recline.arg_types.choices import Choices
from recline.arg_types.flag import Flag
from recline.commands.cli_command import CLICommand
from recline.formatters.table_formatter import TableFormat


# pylint: disable=bad-continuation,unused-argument
def unit_test_command(
    types: Choices.define(["type1", "type2", "type3"]),
    names: List[str],
    optional: str = "foo",
    optional2: bool = False,
    optional3: Flag = False,
) -> TableFormat:
    """This is a test command short description

    This is a test command long description

    Args:
        types: This is the types parameter
        names: This is the names parameter
        optional: This is the optional parameter
        optional2: This is the optional2 parameter
        optional3: This is the optional3 parameter
    """


def test_command_init():
    """Verify that given a function, the CLICommand parses it correctly and contains
    a list of parameters and docstrings for those parameters.
    """

    command = CLICommand(unit_test_command)
    assert command.name == "unit_test_command"
    assert len(command.required_args) == 2
    assert len(command.optional_args) == 3
    assert command.output_formatter is not None
    assert not command.hidden
    assert not command.is_alias


@pytest.mark.parametrize("is_hidden", [True, False])
def test_command_hidden(is_hidden):
    """Verify the command says it is hidden when appropriate"""

    def is_currently_hidden():
        return is_hidden

    def maybe_hidden_command():
        """This command can be hidden"""

    def staticly_hidden_command():
        """This command is always hidden or never hidden"""

    command = CLICommand(maybe_hidden_command, hidden=is_currently_hidden)
    assert command.hidden == is_hidden

    command = CLICommand(staticly_hidden_command, hidden=is_hidden)
    assert command.hidden == is_hidden


@pytest.mark.parametrize("arg_name, expected", [
    ("types", "  -types <type1|type2|type3> This is the types parameter"),
    ("names", "  -names <names> [names ...] This is the names parameter"),
    ("optional", "  -optional <optional> This is the optional parameter\n    Default: foo"),
    ("optional2", "  -optional2 <true|false> This is the optional2 parameter\n    Default: false"),
    ("optional3", "  -optional3 This is the optional3 parameter"),
])
def test_get_arg_help(arg_name, expected):
    """Verify we can generate help text for a given arg"""

    command = CLICommand(unit_test_command)
    for arg in command.required_args:
        if arg.name == arg_name:
            arg_help = command.get_arg_help(arg)
            assert arg_help == expected
    for arg in command.optional_args:
        if arg.name == arg_name:
            arg_help = command.get_arg_help(arg)
            assert arg_help == expected


def test_get_command_help():
    """Verify we can generate help text for the whole command"""

    command = CLICommand(unit_test_command)
    help_text = command.get_command_help()
    assert command.docstring.short_description in help_text
    assert command.docstring.long_description not in help_text
    for arg in command.required_args:
        arg_help = command.get_arg_help(arg)
        assert arg_help in help_text
    for arg in command.optional_args:
        arg_help = command.get_arg_help(arg)
        assert arg_help in help_text


def test_register_start():
    """Verify we can register a single command run run at the start of the applciation"""

    recline.commands.START_COMMAND = None

    @recline.command(atstart=True)
    def start_cmd():
        """This command will run at the start of the application"""

    assert (
        recline.commands.START_COMMAND.func == start_cmd  # pylint: disable=comparison-with-callable
    )

    with pytest.raises(RuntimeError):
        @recline.command(atstart=True)
        def another_start_cmd():  # pylint: disable=unused-variable
            """Can only have one start command"""

    recline.commands.START_COMMAND = None


def test_register_exit():
    """Verify we can register a single command to run at the end of the applciation"""

    recline.commands.EXIT_COMMAND = None

    @recline.command(atexit=True)
    def exit_cmd():
        """This command will run at the exit of the application"""

    assert (
        recline.commands.EXIT_COMMAND.func == exit_cmd  # pylint: disable=comparison-with-callable
    )

    with pytest.raises(RuntimeError):
        @recline.command(atexit=True)
        def another_exit_cmd():  # pylint: disable=unused-variable
            """Can only have one exit command"""

    recline.commands.EXIT_COMMAND = None



"""
Original © NetApp 2024

A test module for the recline.commands.async_command module
"""

import asyncio

import pytest

import recline
from recline.commands.async_command import AsyncCommand


@pytest.mark.usefixtures("clean_jobs")
def test_async_command_run():
    """Verify that if given a coroutine, the AsyncCommand will register and then
    run the coroutine when in the foreground.
    """

    test_result = "This is the answer"

    @recline.command(name="test")
    async def test_coro():  # pylint: disable=unused-variable
        return test_result

    thread = AsyncCommand(recline.commands.COMMAND_REGISTRY["test"])
    job_pid = thread.job_pid
    assert recline.JOBS[job_pid] == thread
    assert recline.NEXT_JOB_PID == job_pid + 1
    thread.start()
    assert thread.foreground() == test_result
    assert job_pid not in recline.JOBS


@pytest.mark.usefixtures("clean_jobs")
def test_async_command_run_background():
    """Verify that if given a coroutine, the AsyncCommand will register and then
    run the coroutine when in the background and we can retrieve the result.
    """

    test_result = "This is the answer"

    @recline.command(name="test")
    async def test_coro():  # pylint: disable=unused-variable
        return test_result

    thread = AsyncCommand(recline.commands.COMMAND_REGISTRY["test"])
    thread.start()
    thread.background()
    # wait until it is complete
    while thread.is_alive():
        pass
    assert thread.result == test_result


@pytest.mark.usefixtures("clean_jobs")
def test_async_command_stop():
    """Verify that we can kill a long running command"""

    i_was_stopped = False
    i_was_started = False

    @recline.command(name="test")
    async def test_coro():  # pylint: disable=unused-variable
        nonlocal i_was_stopped, i_was_started
        try:
            while True:
                i_was_started = True
                await asyncio.sleep(0.001)
        except asyncio.CancelledError:
            i_was_stopped = True

    thread = AsyncCommand(recline.commands.COMMAND_REGISTRY["test"])
    job_pid = thread.job_pid
    thread.start()
    # wait until it is started
    while not i_was_started:
        pass
    thread.stop()
    # wait until it is stopped
    while thread.is_alive():
        pass
    assert i_was_stopped
    assert job_pid not in recline.JOBS


@pytest.mark.usefixtures("clean_jobs")
def test_async_command_exception():
    """Verify that an exception thrown in the background thread is raised to the
    foreground thread when it is foregrounded
    """

    test_result = "This is the exception"

    @recline.command(name="test")
    async def test_coro():  # pylint: disable=unused-variable
        raise RuntimeError(test_result)

    thread = AsyncCommand(recline.commands.COMMAND_REGISTRY["test"])
    job_pid = thread.job_pid
    assert recline.JOBS[job_pid] == thread
    assert recline.NEXT_JOB_PID == job_pid + 1
    thread.start()
    with pytest.raises(RuntimeError, match=test_result):
        assert thread.foreground() == test_result
    assert job_pid not in recline.JOBS



"""
Original © NetApp 2024

This package contains the base command class and the builtin command for all
recline-based applications to use.
"""

COMMAND_REGISTRY = {}
START_COMMAND = None
EXIT_COMMAND = None

class ReclineCommandError(Exception):
    """If there is an error while executing a command, this exception should be
    raised an an error will be printed to the console.
    """

__all__ = ['COMMAND_REGISTRY', 'EXIT_COMMAND', 'START_COMMAND', 'ReclineCommandError']



"""
Original © NetApp 2024

The commands defined here are included with every recline application
"""

from collections import OrderedDict
import curses
from operator import attrgetter
import sys

import recline
from recline.arg_types.choices import Choices
from recline.arg_types.flag import Flag
from recline.arg_types.remainder import Remainder
from recline.commands import ReclineCommandError
from recline.commands.cli_command import command
from recline.commands.man_utils import  generate_help_text


class DebugInterrupt(Exception):
    """An exception class to mark that we were interrupted by a debug request"""


EXIT_COMMAND_CODE = 222
_GROUPNAME = 'Built-in Commands'


@command(group=_GROUPNAME, aliases=['?'], name="help")
def command_help() -> None:
    """Display a list of available commands and their short description"""

    groups = OrderedDict()
    print('Available Commands:')
    print()
    for command_entry in recline.commands.COMMAND_REGISTRY.values():
        if command_entry.is_alias or command_entry.hidden:
            continue

        command_group = command_entry.group if command_entry.group else ""
        if command_group not in groups:
            groups[command_group] = []
        groups[command_group].append(command_entry)

    # print the builtin functions last
    builtins = groups.pop(_GROUPNAME, [])

    for group in sorted(groups.keys()):
        if group:
            print(f"{group}\n{'-' * len(group)}")
        for command_entry in sorted(groups[group], key=attrgetter('name')):
            print(f'{command_entry.name} - {command_entry.docstring.short_description}')

    print()
    print(f"{_GROUPNAME}\n{'-' * len(_GROUPNAME)}")
    for command_entry in sorted(builtins, key=attrgetter('name')):
        if command_entry.is_alias or command_entry.hidden:
            continue
        print(f'{command_entry.name} - {command_entry.docstring.short_description}')


def man_commands(*_, **__):
    """A completer function for returning a list of commands the user might want
    to see the man page for. Don't show alias names to reduce clutter and confusion.
    """

    return [
        name for name, command in recline.commands.COMMAND_REGISTRY.items() if not command.is_alias
    ]


@command(group=_GROUPNAME)
# pylint: disable=too-many-statements
def man(command_name: Remainder.define(completer=man_commands)) -> None:
    """Display the full man page for a given command

    Args:
        command: The name of the command
    """

    command_name = ' '.join(command_name)
    command_class = recline.commands.COMMAND_REGISTRY.get(command_name, None)
    if not command_class:
        print(f'No manual entry for {command_name}')
        return

    def display_man(window):
        curses.curs_set(False)
        max_lines, max_cols = window.getmaxyx()
        help_text = generate_help_text(max_cols, command_class)
        window.clear()
        window.refresh()
        top = 0
        at_bottom = False
        at_top = True

        def _redraw_man_page():
            nonlocal at_bottom, at_top
            window.clear()

            # title
            window.addstr(0, 0, command_name)
            window.addstr(0, (max_cols - len(recline.PROGRAM_NAME)) // 2, recline.PROGRAM_NAME)
            window.addstr(0, max_cols - len(command_name), command_name)
            window.addstr('\n\n')

            index = top
            at_bottom = True
            at_top = top == 0
            while index < len(help_text):
                current_y, _ = window.getyx()
                room_left = max_lines - current_y - 1
                line = help_text[index]
                if '\n' not in line[0]:
                    window.addstr(*line)
                    index += 1
                    continue
                lines = line[0].split('\n')
                if len(lines) > room_left:
                    window.addstr('\n'.join(lines[:room_left]), *line[1:])
                    at_bottom = False
                    break
                window.addstr(*line)
                index += 1
            if at_bottom:
                window.addstr(max_lines - 1, 0, '(END)', curses.A_STANDOUT)

            window.refresh()

        # initial draw
        _redraw_man_page()

        # get key presses
        while True:
            key = window.getch()
            if key == ord('q'):
                break

            if key == curses.KEY_DOWN and not at_bottom:
                top += 1
                # if scrolling down and the "line" we just consumed isn't a full
                # line, then we need to keep going until we consume a \n
                while top < len(help_text) and not help_text[top - 1][0].endswith('\n'):
                    top += 1
            elif key == curses.KEY_UP and not at_top:
                top -= 1
                # if scrolling up and the "line" we're going to consume next time
                # isn't a full line, then we need to move the pointer until it is
                while top > 0 and not help_text[top - 1][0].endswith('\n'):
                    top -= 1

            _redraw_man_page()

    curses.wrapper(display_man)


@command(group=_GROUPNAME, aliases=['quit', 'q'], name="exit")
def exit_command(abort_jobs: Flag = False) -> None:
    """Exit the application"""

    if not abort_jobs and recline.JOBS:
        answer = input("There are backgrounded jobs. Are you sure you want to quit? ")
        if answer.lower().startswith("y"):
            abort_jobs = True
        else:
            return

    if abort_jobs:
        # clean up the jobs before we call sys.exit. This prevents an issue where
        # a traceback is printed due to some cleanup code in async tasks:
        # > q -abort-jobs
        # Exception ignored in: <object repr() failed>
        # Traceback (most recent call last):
        # File "/usr/software/pkgs/Python-3.5.2/lib/python3.5/asyncio/tasks.py", line 85, in __del__
        # AttributeError: 'NoneType' object has no attribute '_PENDING'
        for job in recline.JOBS.values():
            job.stop(dont_delete=True)
        recline.JOBS = {}

    # now we should be safe to exit
    sys.exit(EXIT_COMMAND_CODE)


@command(group=_GROUPNAME, hidden=True)
def debug() -> None:
    """Drop into the Python debugger for development purposes"""

    try:
        import pudb  # pylint: disable=import-outside-toplevel
        # avoid pudb printing out an ugly message
        def no_native_int_support(*_x, **_y):
            pass

        pudb.set_interrupt_handler = no_native_int_support

        # the program pauses after this line
        pudb.set_trace()
    except ImportError:
        import pdb  # pylint: disable=import-outside-toplevel
        pdb.set_trace()
    finally:
        # remind the user that they can pause the program at any time
        print(r'Press ctrl+\ to break back to the debugger')


@command(group=_GROUPNAME)
def fg(job: Choices.define(  # pylint: disable=invalid-name
        available_choices=lambda: list(recline.JOBS.keys()), data_type=int,
    )=None) -> None:
    """Bring a job to the foreground

    If the job has already completed, then its results will be printed to the
    console. If the job has not yet finished, then the application will wait for
    it to finish before printing the results.

    Args:
        job: The identifier of the job to track. If not provided, then the most
            recent job will be tracked.
    """

    if not job:
        try:
            job = sorted(recline.JOBS.keys())[-1]
        except IndexError:
            raise ReclineCommandError("No running jobs found")
    elif job not in recline.JOBS:
        raise ReclineCommandError(f"Could not find a running job for {job}")

    thread = recline.JOBS[job]
    result = thread.foreground()
    if thread.command.output_formatter:
        thread.command.output_formatter.format_output(result)



"""
Original © NetApp 2024

A Remainder type allows the CLI command writer to specify that an argument should
consume all of the rest of the input from the CLI and that it will be interpreted
inside of the command handler function.

@recline.command
def search(query: Remainder) -> None:
    # The user might use this command like "search stuff I want to know about"
    # and the query parameter would be filled with the value "stuff I want to know about"
"""

import argparse

from recline.arg_types.recline_type import ReclineType


class Remainder(ReclineType):
    """The Remainder type will capture all of the remaining input from the user
    into a single parameter, even if that input contained spaces which would
    normally be considered multiple values
    """

    @staticmethod
    def define(completer=None):
        """Allows a Remainder type to be defined with its own completer function"""

        class _Remainder(Remainder):
            metavar = '<value>'

            def __init__(self):
                if completer:
                    self.completer = completer
                super().__init__()

        return _Remainder

    def nargs(self):
        return argparse.REMAINDER



"""
Original © NetApp 2024

A RangedInt type allows the CLI command writer to specify an integer that is only
valid within a certain range. If the user provides a value outside of that range,
then the parameter validation will fail and they will receive an error message.

@recline.command(name="bake cake")
def bake(minutes: RangedInt.define(min=15, max=60)) -> None:
    # If the user tried to bake the cake for less than 15 minutes or more than
    # 60 minutes, then this body wouldn't be called and they would get an error.
"""

from recline.arg_types.recline_type import ReclineType
from recline.arg_types.recline_type_error import ReclineTypeError


class RangedInt(ReclineType):
    """The RangedInt type allows a parameter which is an into to be constrained
    in which values are in the valid range.
    """

    # pylint: disable=bad-continuation
    @staticmethod
    def define(
        min: int = None,  # pylint: disable=redefined-builtin
        max: int = None,  # pylint: disable=redefined-builtin
    ) -> "_RangedInt":
        """A `recline.commands.types.RangedInt` is a way to annotate an integer with
        a min and max value.

        Prior to being passed to your command, the arguments of this type will
        be validated according to the parameters specified. For example, if a
        minimum is given, then the argument must be less than the minimum before
        your command will be called with it.

        Args:
            name: This is the name of the parameter and is required for building
                the help text of the command.
            min: If provided, the user-supplied argument will be validated to be
                and integer >= this value.
                argument will be enforced.
            max: If provided, the user-supplied argument will be validated to be
                and integer <= this value.
        """

        min_val = min
        max_val = max

        class _RangedInt(RangedInt):
            metavar = '<int>'

            range_str = ''
            if min_val is not None and max_val is not None:
                range_str = '{%s-%s}' % (min_val, max_val)
            elif min_val is not None:
                range_str = '{%s-inf}' % min_val
            elif max_val is not None:
                range_str = '{-inf-%s}' % max_val
            metavar = f'<int{range_str}>'

            def validate(self, arg):
                try:
                    int_val = int(arg)
                    if min_val is not None and int_val < min_val:
                        raise ValueError()
                    if max_val is not None and int_val > max_val:
                        raise ValueError()
                    return int_val
                except ValueError:
                    raise ReclineTypeError(
                        f'"{arg}" is not an integer in the range {_RangedInt.range_str}.'
                    )

        return _RangedInt



"""
Original © NetApp 2024

A test module for the recline.arg_types.recline_type module
"""

import argparse
from contextlib import ExitStack as does_not_raise

import pytest

from recline.arg_types.recline_type import UniqueParam


@pytest.mark.parametrize('user_input, expectation', [
    ([], does_not_raise()), (['-arg', 'foo'], does_not_raise()),
    (['-arg2', 'foo'], does_not_raise()),
    (['-arg', 'foo', '-arg', 'bar'], pytest.raises(SystemExit)),
    (['-arg', 'foo', '-arg2', 'bar'], does_not_raise()),
    (['-arg2', 'foo', '-arg2', 'bar'], pytest.raises(SystemExit)),

])
def test_unique_param(user_input, expectation):
    """Verify we get an error if we try to parse input that duplicates a unique
    parameter in our definition
    """

    parser = argparse.ArgumentParser()
    parser.add_argument('-arg', action=UniqueParam)
    parser.add_argument('-arg2', action=UniqueParam)
    with expectation:
        parser.parse_args(user_input)



"""
Original © NetApp 2024

A test module for the recline.repl.completer module
"""

import pytest

import recline
from recline.commands.cli_command import command
from recline.repl.completer import CommandCompleter, match_command_hook


@pytest.mark.parametrize("prompt, current_input, matches, expected_output", [
    (
        "> ", "input", [],
        "\n\n> \r> ",
    ),
    (
        ">>> ", "part", ["part of", "partial"],
        "\npart of  partial  \n>>> \r>>> ",
    ),
    (
        "> ", "p", ["part of", "partial", "partner", "pool", "pony"],
        "\npart of  partial  partner  \npool     pony     \n> \r> ",
    ),
])
# pylint: disable=too-many-arguments
def test_match_command_hook(prompt, current_input, matches, expected_output, monkeypatch, capsys):
    """Verify our match hook for readline's set_completion_display_matches_hook()
    prints out the right output
    """

    monkeypatch.setattr(recline, "PROMPT", prompt)
    match_command_hook(current_input, matches)
    captured = capsys.readouterr()
    assert captured.out == expected_output


@pytest.mark.parametrize("partial, possibles, exact, match", [
    ("cake show -cake 1", ["cake show", "cake make", "cake make again"], False, "cake show"),
    ("cake make", ["cake show", "cake make", "cake make again"], False, None),
    ("cake make", ["cake show", "cake make", "cake make again"], True, "cake make"),
    ("help", ["cake show", "cake make", "cake make again"], True, None),
])
def test_command_completer_get_command_name(partial, possibles, exact, match):
    """Verify we can complete a command name given a partial and a list of possibles
    where only one possible matches the partial.
    """

    assert CommandCompleter.get_command_name(partial, possibles, exact=exact) == match


@pytest.mark.parametrize("partial, index, commands, expected_output", [
    ("", 0, ["cake show", "cake make", "cake make again", "cake eat"], "cake eat "),
    ("cak", 0, ["cake show", "cake make", "cake make again", "cake eat"], "cake eat "),
    ("cak", 1, ["cake show", "cake make", "cake make again", "cake eat"], "cake make "),
    ("cak", 2, ["cake show", "cake make", "cake make again", "cake eat"], "cake make again "),
    ("cak", 3, ["cake show", "cake make", "cake make again", "cake eat"], "cake show "),
    ("cake show", 0, ["cake show", "cake make", "cake make again", "cake eat"], "cake show "),
    ("cake show ", 0, ["cake show", "cake make", "cake eat"], "cake show -arg "),
    ("cake show -a", 0, ["cake show", "cake make", "cake eat"], "cake show -arg "),
    ("cake show -arg 1 ", 0, ["cake show", "cake make", "cake eat"], None),
    ("cak", 0, ["cake", "cake show"], "cake "),
    ("cak", 1, ["cake", "cake show"], "cake show "),
    ("cake", 0, ["cake", "cake show"], "cake "),
    ("cake -", 0, ["cake", "cake show"], "cake -arg "),
    ("help", 0, ["cake show", "cake make", "cake make again", "cake eat"], None),
])
def test_command_completer_completer(partial, index, commands, expected_output):
    """Verify the command completer will iterate through all possible matches
    of the user's current text.
    """

    # this is why global state is dangerous folks...
    recline.commands.COMMAND_REGISTRY = {}

    for app_command in commands:
        @command(name=app_command)
        def ut_command(arg: str):  # pylint: disable=unused-variable,unused-argument
            pass

    completer = CommandCompleter(recline.commands.COMMAND_REGISTRY)
    assert completer.completer(partial, index) == expected_output



"""
Original © NetApp 2024

This module contains the abstract base class for all recline custom types
"""

from abc import ABC
import argparse
from typing import Any, List, Optional, Union


class UniqueParam(argparse.Action):  # pylint: disable=too-few-public-methods
    """Utility class that throws an error if more than one occurrence of param is found."""

    def __call__(self, parser, namespace, values, option_string=None):
        current_value = getattr(namespace, self.dest)
        # If the corresponding namespace attribute is not the default value or None raise error
        if current_value != self.default and current_value is not None:
            parser.error(f"Duplicate parameter \"{option_string}\"")
        # Update the namespace attribute with the given value
        setattr(namespace, self.dest, values)


class ReclineType(ABC):
    """The base class for all other custom types in recline. Default values for
    argparse are defined within each type.
    """

    metavar = "<value>"
    action = UniqueParam

    def __init__(self):
        self.arg_name = None

    def choices(self, eager=False) -> Optional[List[Any]]:  # pylint: disable=unused-argument,no-self-use
        """If an argument has a set list of choices that can be provided for it,
        then the type can specify to only allow that list
        """

        return None

    def completer(self, *args, **kwargs) -> List[Any]:  # pylint: disable=unused-argument,no-self-use
        """The completer function should return a list of values that are valid
        for the argument. This function will usually be implemented such that it
        is dynamic based on some API call or some current application state. If
        it is a static set of choices, it would be easier to use the choices
        method instead.
        """

        return [None]

    def nargs(self) -> Optional[Union[int, str]]:  # pylint: disable=no-self-use
        """The number of arguments that this type can accept. See the argparse
        documentation for details: https://docs.python.org/3/library/argparse.html#nargs
        """

        return None

    def validate(self, arg: str) -> Any:  # pylint: disable=no-self-use
        """The validate function has two uses. First, a type should verify that
        the user's input matches the data that the type accepts. Second, it should
        transform the data if necessary. For example, it might try to cast a string
        to an integer and if it can, return the integer value.

        If validation fails, a ReclineTypeError should be raised with an appropriate
        message for the user.
        """

        return arg



"""
Original © NetApp 2024

A test module for the recline.commands.builtin_commands module
"""

import pdb
import pudb
import pytest

import recline
from recline.commands import builtin_commands
from recline.commands.cli_command import CLICommand


@pytest.mark.parametrize("commands, expected_commands, expected_groups", [
    (
        [{"name": "command1", "group": "group1", "is_alias": False, "hidden": False}],
        ["command1"],
        ["group1"],
    ),
    (
        [{"name": "command1", "group": "group1", "is_alias": False, "hidden": True}],
        [],
        [],
    ),
    (
        [
            {"name": "command1", "group": "group1", "is_alias": False, "hidden": False},
            {"name": "c1", "group": "group1", "is_alias": True, "hidden": False},
        ],
        ["command1"],
        ["group1"],
    ),
    (
        [
            {"name": "command1", "group": "group1", "is_alias": False, "hidden": False},
            {"name": "command2", "group": "group2", "is_alias": False, "hidden": False},
            {"name": "command3", "group": "group1", "is_alias": True, "hidden": False},
            {"name": "command4", "group": "group2", "is_alias": False, "hidden": True},
        ],
        ["command1", "command2"],
        ["group1", "group2"],
    ),
    (
        [
            {
                "name": "builtincommand",
                "group": builtin_commands._GROUPNAME,  # pylint: disable=protected-access
                "is_alias": False, "hidden": False,
            },
            {"name": "othercommand", "group": "group1", "is_alias": False, "hidden": False},
            {"name": "othercommand2", "group": "group1", "is_alias": False, "hidden": True},
        ],
        ["builtincommand", "othercommand"],
        [builtin_commands._GROUPNAME, "group1"],  # pylint: disable=protected-access
    ),
])
def test_help_command(commands, expected_commands, expected_groups, capsys):
    """Verify that the help command prints out all of the available commands and
    none of their aliases or hidden commands.
    """

    for command in commands:
        recline.commands.COMMAND_REGISTRY[command["name"]] = CLICommand(lambda: None, **command)

    builtin_commands.command_help()
    captured = capsys.readouterr()
    for command in expected_commands:
        assert command in captured.out
    for group in expected_groups:
        assert group in captured.out

    for command in commands:
        if command["name"] not in expected_commands:
            assert command["name"] not in captured.out
        if command["group"] not in expected_groups:
            assert command["group"] not in captured.out


def test_exit_command():
    """Verify we exit the application"""

    with pytest.raises(SystemExit):
        builtin_commands.exit_command()


@pytest.mark.parametrize('use_pudb', [True, False])
def test_debug_command(use_pudb, monkeypatch, capsys):
    """Verify that we try to set a trace point"""

    set_a_trace = False

    def mock_set_trace():
        nonlocal set_a_trace
        set_a_trace = True

    def raise_set_trace():
        raise ImportError('Non-existent for UT pretend')

    if use_pudb:
        monkeypatch.setattr(pudb, 'set_trace', mock_set_trace)
    else:
        monkeypatch.setattr(pudb, 'set_trace', raise_set_trace)
    monkeypatch.setattr(pdb, 'set_trace', mock_set_trace)

    builtin_commands.debug()
    assert set_a_trace
    captured = capsys.readouterr()
    assert 'break back to the debugger' in captured.out



"""
Original © NetApp 2024

A test module for the recline.arg_types.positional module
"""

import argparse
from contextlib import ExitStack as does_not_raise

import pytest

from recline.arg_types.recline_type_error import ReclineTypeError
from recline.arg_types.positional import Positional


@pytest.mark.parametrize("user_input, expectation", [
    ("my_dir", does_not_raise()), ("", pytest.raises(ReclineTypeError)),
])
def test_positional(user_input, expectation):
    """Verify the Positional type will assert the user provided something"""

    def path_completer(*_, **__):
        """This completer would be called at runtime if the user pressed tab"""
        return []

    default_positional = Positional()
    assert default_positional.validate(user_input) == user_input
    assert default_positional.nargs() is None
    assert default_positional.completer()[0] is None

    parser = argparse.ArgumentParser()
    parser.add_argument('path', type=Positional.define(completer=path_completer)().validate)

    with expectation:
        parsed_args = parser.parse_args([user_input])
        assert parsed_args.path == user_input


@pytest.mark.parametrize("user_input, data_type, expectation", [
    ("my_val", str, does_not_raise()), ("my_val", int, pytest.raises(ReclineTypeError)),
])
def test_positional_type(user_input, data_type, expectation):
    """Verify the Positional type will assert the user provided a correct type"""

    parser = argparse.ArgumentParser()
    parser.add_argument('value', type=Positional.define(data_type=data_type)().validate)
    with expectation:
        parsed_args = parser.parse_args([user_input])
        assert parsed_args.value == user_input



"""
Original © NetApp 2024

A test module for the recline.arg_types.ranged_int module
"""

from contextlib import ExitStack as does_not_raise

import pytest

from recline.arg_types.recline_type_error import ReclineTypeError
from recline.arg_types.ranged_int import RangedInt


@pytest.mark.parametrize("min_val, max_val, user_input, expectation", [
    (2, 10, 5, does_not_raise()), (0, None, 100, does_not_raise()),
    (None, 100, -50, does_not_raise()), (None, None, 100, does_not_raise()),
    (2, 10, 1, pytest.raises(ReclineTypeError)), (0, None, -1, pytest.raises(ReclineTypeError)),
    (None, 100, 105, pytest.raises(ReclineTypeError)),
    (2, 10, "foo", pytest.raises(ReclineTypeError)),
])
def test_choices(min_val, max_val, user_input, expectation):
    """Verify the RangedInt type will assert the user provided something in the
    specified range"""

    ranged_int = RangedInt.define(min=min_val, max=max_val)()
    with expectation:
        ranged_int.validate(user_input)



"""
Original © NetApp 2024

This package contains the various custom type definitions for CLI command parameters.
If you are defining a CLI command function that takes input with restrictions,
you may want to use a type annotation to allow recline to do the checking for you
before your function ever gets invoked. For example:

from recline.arg_types import Choices

@recline.command
def commit(branch: Choices.define(['dev', 'beta', 'stable'])) -> None:
    # the function can now assume that branch has already been validated to be
    # one of the options above before we got here
"""



"""
Original © NetApp 2024
"""

import os
import sys

PROGRAM_NAME = None
PROMPT = '> '
NEXT_JOB_PID = 1
JOBS = {}

from recline.commands.cli_command import command  # pylint: disable=wrong-import-position
from recline.repl.shell import relax  # pylint: disable=wrong-import-position

__all__ = ['PROGRAM_NAME', 'PROMPT', 'command', 'relax']



"""
Original © NetApp 2024

A test module for the recline.arg_types.choices module
"""

from contextlib import ExitStack as does_not_raise

import pytest

from recline.arg_types.choices import Choices
from recline.arg_types.recline_type_error import ReclineTypeError


@pytest.mark.parametrize("valid_choices, inexact, user_choice, expectation", [
    (["chocolate", "vanilla"], False, "chocolate", does_not_raise()),
    (["chocolate", "vanilla"], False, "strawberry", pytest.raises(ReclineTypeError)),
    ([], False, "anything", pytest.raises(ReclineTypeError)),
    (["chocolate", "vanilla"], True, "cho*", does_not_raise()),
])
def test_choices(valid_choices, inexact, user_choice, expectation):
    """Verify the Choices type will be picky about which values are allowed"""

    choices = Choices.define(valid_choices, inexact=inexact)()
    assert choices.completer() == valid_choices
    with expectation:
        choices.validate(user_choice)


@pytest.mark.parametrize("valid_choices, cache, user_choice, expectation, func_access", [
    (["chocolate", "vanilla"], True, "chocolate", does_not_raise(), 1),
    (["chocolate", "vanilla"], False, "chocolate", does_not_raise(), 3),
    (["chocolate", "vanilla"], True, "strawberry", pytest.raises(ReclineTypeError), 1),
    ([], "anything", False, pytest.raises(ReclineTypeError), 2),
])
def test_choices_callable(valid_choices, cache, user_choice, expectation, func_access):
    """Verify the Choices type will be picky about which values are allowed when
    the input to the class is a callable.
    """

    times_called = 0

    def get_choices():
        nonlocal times_called
        times_called += 1
        return valid_choices

    choices = Choices.define(get_choices, cache_choices=cache)()
    assert choices.choices() is None
    assert choices.choices(eager=True) == valid_choices
    assert choices.completer() == valid_choices
    with expectation:
        choices.validate(user_choice)
        assert times_called == func_access



"""
Original © NetApp 2024

This module holds some utility functions used as part of the man command to format
text from CLI commands into consistent man pages that respond to the terminal
width.
"""

import curses

from recline.arg_types.positional import Positional
from recline.arg_types.remainder import Remainder
from recline.commands.cli_command import get_annotation_type


def wrapped_string(text, screen_width, prefix=0):
    """This function will take a string and make sure it can fit within the
    given screen_width.

    If the string is too long to fit, it will be broken on word boundaries
    (specifically the ' ' character) if it can or the word will be split with
    a '-' character and the second half moved to the next line.

    If a prefix is given, the line(s) will be prefixed with that many ' '
    characters, including any wrapped lines.

    If the given string includes embedded newline characters, then each line
    will be evaluated according to the rules above including breaking on word
    boundaries and injecting a prefix.
    """

    if not text:
        return ''
    new_text = ''

    # if we have multiple paragraphs, then wrap each one as if it were a single line
    lines = text.split('\n')
    if len(lines) > 1:
        for index, line in enumerate(lines):
            if index > 0:
                new_text += ' ' * prefix
            new_text += wrapped_string(line, screen_width, prefix=prefix) + '\n'
        return new_text.rstrip()

    if len(text) + prefix < screen_width:
        return text
    words = text.split(' ')
    current_line = ''
    for word in words:
        if prefix + len(current_line) + len(word) + 1 < screen_width:
            # if word fits on line, just add it
            current_line += word + ' '
        else:
            space_left = screen_width - (prefix + len(current_line))
            if space_left < 3 or len(word) - space_left < 3:
                # if not much room, move whole word to the next line
                new_text += f'{current_line.rstrip()}\n'
                current_line = f"{' ' * prefix}{word} "
            else:
                # split the word across lines with a hyphen
                current_line += word[:space_left - 1] + '-'
                new_text += current_line.rstrip() + "\n"
                current_line = ' ' * prefix + word[space_left - 1:] + ' '
    new_text += current_line.rstrip()
    return new_text


def generate_help_text(screen_width, command_class):
    """Generates lines of help text which are formatted using the curses library.
    The final document resembles a typical Linux-style manpage. See here:
    https://www.tldp.org/HOWTO/Man-Page/q3.html
    """

    # generate styled man page, one section at a time
    help_text = []
    indent = '     '

    # command name and short description
    help_text.append(('NAME\n', curses.A_BOLD))
    help_text.append((indent,))
    help_text.append((command_class.name, curses.A_BOLD))
    help_text.append((' -- ',))
    description = wrapped_string(
        command_class.docstring.short_description, screen_width,
        prefix=(len(command_class.name) + len(indent) + 4),
    )
    for line in description.split('\n'):
        help_text.append((f'{line}\n',))
    help_text.append(('\n',))

    # command usage details
    help_text.append(('SYNOPSIS\n', curses.A_BOLD))
    help_text.append((indent,))
    description = wrapped_string(
        command_class.get_command_usage(), screen_width, prefix=len(indent),
    )
    for line in description.split('\n'):
        help_text.append((f'{line}\n',))
    help_text.append(('\n',))

    # command detailed description
    if command_class.docstring.long_description:
        help_text.append(('DESCRIPTION\n', curses.A_BOLD))
        help_text.append((indent,))
        description = wrapped_string(
            command_class.docstring.long_description, screen_width,
            prefix=len(indent),
        )
        for line in description.split('\n'):
            help_text.append((f'{line}\n',))
        help_text.append(('\n',))

    # each command parameter with description, constraints, and defaults
    if command_class.docstring.params:
        def print_arg(arg):
            meta = command_class.get_arg_metavar(arg)
            description = command_class.get_arg_description(arg, indent=None)
            annotation_type = get_annotation_type(arg)
            positional = '' if issubclass(annotation_type, (Remainder, Positional)) else '-'
            arg_name = '' if issubclass(annotation_type, Positional) else f'{arg.name} '
            prefix = f'{indent}  {positional}{arg_name}{meta} '
            help_text.append((prefix,))
            description = wrapped_string(
                description, screen_width, prefix=len(prefix)
            )
            for line in description.split('\n'):
                help_text.append((f'{line}\n',))
            help_text.append(('\n'),)
        help_text.append(('OPTIONS\n', curses.A_BOLD))
        if command_class.required_args:
            help_text.append((indent,))
            help_text.append(('Required:\n', curses.A_UNDERLINE))
        for arg in command_class.required_args:
            print_arg(arg)
        if command_class.optional_args:
            help_text.append((indent,))
            help_text.append(('Optional:\n', curses.A_UNDERLINE))
        for arg in command_class.optional_args:
            print_arg(arg)

    # each command example
    if command_class.docstring.examples:
        help_text.append(('EXAMPLES\n', curses.A_BOLD))
        for example in command_class.docstring.examples:
            prefix = indent + '  '
            help_text.append((indent,))
            help_text.append((f'{example.name}:', curses.A_UNDERLINE))
            help_text.append((f'\n{prefix}',))
            description = wrapped_string(
                example.description, screen_width, prefix=len(prefix),
            )
            for line in description.split('\n'):
                help_text.append((f'{line}\n',))

    return help_text



"""
Original © NetApp 2024

This is the implementation of an async command for the recline library. It allows
for a command to be run in the foreground or background.
"""

import asyncio
from contextlib import contextmanager
import io
import os
import signal
import sys
from threading import Thread
import time
from typing import Any

import recline


class CommandBackgrounded(Exception):
    """When a command is moved from the foreground to the background, this exception
    is raised to signal the shell.
    """

    def __init__(self, job_pid, *args, **kwargs):
        self.job_pid = job_pid
        super().__init__(*args, **kwargs)


class CommandCancelled(Exception):
    """When a command is cancelled, this is exception is raised to signal the shell"""

    def __init__(self, job_pid, *args, **kwargs):
        self.job_pid = job_pid
        super().__init__(*args, **kwargs)


# pylint: disable=too-many-instance-attributes
class AsyncCommand(Thread):
    """This is a custom Thread class meant to wrap an asynchronous recline command.

    It provides the ability to gather a result as well as put the thread in the
    foreground or the background of the main REPL thread.
    """

    def __init__(self, command, *args, **kwargs):
        super().__init__()

        self.result = None
        self.exception = None
        self.command = command
        self.job_pid = recline.NEXT_JOB_PID
        recline.JOBS[self.job_pid] = self
        recline.NEXT_JOB_PID += 1

        self._args = args
        self._kwargs = kwargs
        self._loop = asyncio.new_event_loop()
        self._task = None
        self._wait = True
        self._killed = False
        self._orig_sigint_handler = None
        self._orig_sigstp_handler = None

    def run(self) -> None:
        """Start the command in an asyncio loop. When it's complete, the result
        will be recorded in this thread
        """

        if sys.platform != "win32":
            self._orig_sigstp_handler = signal.getsignal(signal.SIGTSTP)
        self._orig_sigint_handler = signal.getsignal(signal.SIGINT)
        asyncio.set_event_loop(self._loop)
        self._task = asyncio.ensure_future(self.command.func(*self._args, **self._kwargs))
        try:
            self.result = self._loop.run_until_complete(self._task)
        except Exception as ex:  # pylint: disable=broad-except
            self.exception = ex

    def stop(self, dont_delete=False) -> None:
        """Kill the command and return control to the main thread"""

        self._task.cancel()
        signal.signal(signal.SIGINT, self._orig_sigint_handler)
        self._killed = True
        if dont_delete:
            return
        del recline.JOBS[self.job_pid]

    def background(self) -> None:
        """Unblock the calling thread and put this thread into the background"""

        self._wait = False
        if sys.platform != "win32":
            signal.signal(signal.SIGTSTP, self._orig_sigstp_handler)

    def foreground(self) -> Any:
        """Block the calling thread until this thread is complete and return the result

        Also setup up a signal handler for SIGTSTP so that pressing ctrl+z will
        put this thread in the background
        """

        self._wait = True
        with self._manipulate_signals(), set_terminal_echo(enabled=False):
            while self._wait and self.is_alive():
                time.sleep(.01)
            if self._killed:
                raise CommandCancelled(self.job_pid)
            if not self._wait:
                raise CommandBackgrounded(self.job_pid)
            if self.exception:
                del recline.JOBS[self.job_pid]
                raise self.exception
            return recline.JOBS.pop(self.job_pid).result

    @contextmanager
    def _manipulate_signals(self):
        if sys.platform != "win32":
            signal.signal(
                signal.SIGTSTP,
                lambda signum, frame: self.background(),
            )
        signal.signal(
            signal.SIGINT,
            lambda signum, frame: self.stop(),
        )
        yield
        if sys.platform != "win32":
            signal.signal(signal.SIGTSTP, self._orig_sigstp_handler)
        signal.signal(signal.SIGINT, self._orig_sigint_handler)


@contextmanager
def set_terminal_echo(enabled=True):
    """ Code from https://docs.python.org/2/library/termios.html """

    try:
        import termios
    except ImportError:
        # must be a Windows system. There's no terios module there, so there's
        # nothing we can do here
        yield
        return

    # in something like pytest, stdin might not have a fileno
    try:
        stdin = sys.stdin.fileno()
    except io.UnsupportedOperation:
        yield
        return

    # if we're not running interactively, then there's no need (or ability) to
    # change the echo status
    if not os.isatty(stdin):
        yield
        return

    old = termios.tcgetattr(stdin)
    new = termios.tcgetattr(stdin)
    if not enabled:
        new[3] = new[3] & ~termios.ECHO
    else:
        new[3] = new[3] | termios.ECHO
    try:
        termios.tcsetattr(stdin, termios.TCSADRAIN, new)
        yield
    finally:
        termios.tcsetattr(stdin, termios.TCSADRAIN, old)



"""
Original © NetApp 2024

A test module for the recline.commands.man_utils module
"""

from inspect import Parameter

import pytest

from recline.commands import man_utils
from recline.commands.cli_command import CLICommand
from recline.vendor.docstring_parser.parser.common import DocstringMeta


@pytest.mark.parametrize("text, screen_width, prefix, output", [
    ("", 5, 0, ""), ("Hi", 5, 0, "Hi"), ("Hello", 10, 0, "Hello"),
    (
        "Hello there, fellow tester", 10, 0,
        "Hello\n"
        "there, fe-\n"
        "llow\n"
        "tester"
    ),
    (
        "Hello there,\nfellow tester", 10, 0,
        "Hello\n"
        "there,\n"
        "fellow te-\n"
        "ster"
    ),
    (
        "Hello there,\n\nfellow tester", 10, 0,
        "Hello\n"
        "there,\n\n"
        "fellow te-\n"
        "ster"
    ),
    (
        "Hello there, fellow tester", 9, 0,
        "Hello th-\n"
        "ere,\n"
        "fellow\n"
        "tester"
    ),
    (
        "Hello there, fellow tester", 12, 3,
        "Hello th-\n"
        "   ere,\n"
        "   fellow\n"
        "   tester"
    ),
])
def test_wrapped_string(text, screen_width, prefix, output):
    """Verify that we can wrap a string given a screen width. We should do our
    best to break on word boundaries when it makes sense and otherwise we'll
    hyphenate.
    """

    assert man_utils.wrapped_string(text, screen_width, prefix=prefix) == output


@pytest.mark.parametrize("command_name, args, kwargs, short_doc, long_doc, examples", [
    (
        "test command", ["foo"], {"bar": None}, "this is a test command",
        "this is a long desc", {"example1": "example1 text", "example2": "example2 text"},
    ),
    (
        "testcommand", ["foo", "bar"], {}, "this is a test command", "", {},
    ),
    (
        "testcommand", [], {}, "this is a test command", "", {},
    ),
])
# pylint: disable=too-many-arguments
def test_generate_help_text(command_name, args, kwargs, short_doc, long_doc, examples):
    """Verify that we generate an expected man page for a given command. I'm not
    going to verify all the formatting here, just the basics. Makes sure that
    all the pieces make it to the page and that it doesn't blow up. Format testing
    will be done visually which I think is acceptable.
    """

    command_class = CLICommand(lambda: None, name=command_name)
    command_class.name = command_name
    command_class.docstring.short_description = short_doc
    command_class.docstring.long_description = long_doc
    command_class.required_args = [
        Parameter(arg, Parameter.POSITIONAL_ONLY) for arg in args
    ]
    command_class.optional_args = [
        Parameter(arg, Parameter.KEYWORD_ONLY, default=val)
        for arg, val in kwargs.items()
    ]
    if args or kwargs:
        command_class.docstring.meta = [DocstringMeta(["arg"], "")]
    for label, description in examples.items():
        command_class.docstring.meta.append(
            DocstringMeta(["examples", label], description)
        )

    help_text = man_utils.generate_help_text(1000, command_class)
    text_pieces = [spec[0] for spec in help_text]
    help_text = ''.join(text_pieces)
    assert command_name in help_text
    for arg in args:
        assert arg in help_text
    for kwarg in kwargs.keys():
        assert kwarg in help_text
    assert short_doc in help_text
    assert long_doc in help_text
    for label, description in examples.items():
        assert label in help_text
        assert description in help_text
